#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.device.dummy` class.
"""

from PIL import Image

from luma.core.render import canvas
from luma.core.device import dummy

import baseline_data
from helpers import get_reference_image, assert_identical_image


def test_capture_noops():
    device = dummy()
    # All these should have no effect
    device.hide()
    device.show()
    device.cleanup()
    device.contrast(123)
    device.command(1, 2, 4, 4)
    device.data([1, 2, 4, 4])


def test_portrait():
    img_path = get_reference_image('portrait.png')

    with open(img_path, 'rb') as p:
        reference = Image.open(p)

        device = dummy(rotate=1)

        # Use the same drawing primitives as the demo
        with canvas(device) as draw:
            baseline_data.primitives(device, draw)

        assert_identical_image(reference, device.image, img_path)


def test_dither():
    img_path = get_reference_image('dither.png')

    with open(img_path, 'rb') as p:
        reference = Image.open(p)
        device = dummy(mode="1")

        with canvas(device, dither=True) as draw:
            draw.rectangle((0, 0, 64, 32), fill="red")
            draw.rectangle((64, 0, 128, 32), fill="yellow")
            draw.rectangle((0, 32, 64, 64), fill="blue")
            draw.rectangle((64, 32, 128, 64), fill="white")

        assert_identical_image(reference, device.image, img_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
TCA9548A I2C Bus Scanner

Based on: https://learn.adafruit.com/adafruit-tca9548a-1-to-8-i2c-multiplexer-breakout?view=all
Assumes the multiplexer is on address 0x70
"""

import smbus2

TCA_ADDR = 0x70


def mux_select(bus, tca_port):
    assert 0 <= tca_port <= 7
    bus.write_byte(TCA_ADDR, 1 << tca_port)


def scan():
    port = 1   # Change to port 0 for orig Raspberry Pi, Orange Pi, etc
    bus = smbus2.SMBus(port)

    print("Scanning I2C devices on TCA984A Multiplexer")

    for tca_port in range(8):
        print("TCA Port {0}".format(tca_port))
        mux_select(bus, tca_port)

        for addr in range(0x03, 0x78):
            if addr == TCA_ADDR:
                continue

            try:
                bus.read_byte(addr)
                print("Found I2C 0x{0:02X}".format(addr))
            except:
                pass

    print("\ndone")
    bus.close()


if __name__ == "__main__":
    scan()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2020 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.device.parallel_device` class.
"""

from unittest.mock import Mock, call
from luma.core.device import parallel_device


def test_4bit():
    serial = Mock(unsafe=True)
    serial._bitmode = 4
    serial._pulse_time = 1
    pd = parallel_device(serial_interface=serial)

    pd.command(0x10, 0x11, 0xff)
    pd.data([0x41, 0x42, 0x43])

    comm = call(0x01, 0x00, 0x01, 0x01, 0x0f, 0x0f)
    data = call([0x04, 0x01, 0x04, 0x02, 0x04, 0x03])
    serial.command.assert_has_calls([comm])
    serial.data.assert_has_calls([data])


def test_8bit():
    serial = Mock(unsafe=True)
    serial._bitmode = 8
    serial._pulse_time = 1
    pd = parallel_device(serial_interface=serial)

    pd.command(0x10, 0x11, 0xff)
    pd.data([0x41, 0x42, 0x43])

    comm = call(0x10, 0x11, 0xff)
    data = call([0x41, 0x42, 0x43])
    serial.command.assert_has_calls([comm])
    serial.data.assert_has_calls([data])



# -*- coding: utf-8 -*-
# Copyright (c) 2014-2020 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Different implementation strategies for framebuffering
"""

from math import sqrt
from PIL import ImageChops, ImageDraw


class diff_to_previous(object):
    """
    Compare the current frame to the previous frame and tries to calculate the
    differences: this will either yield nothing for a perfect match or else
    the iterator will yield one or more tuples comprising of the image part that
    changed along with the bounding box that describes the areas that are different,
    up to the size of the entire image.

    The image data for the difference is then be passed to a device for
    rendering just those small changes. This can be very quick for small screen
    updates, but suffers from variable render times, depending on the changes
    applied. The :py:class:`luma.core.sprite_system.framerate_regulator` may be
    used to counteract this behavior however.

    :param num_segments: The number of segments to partition the image into. This
        generally must be a square number (1, 4, 9, 16, ...) and must be able to
        segment the image entirely in both width and height. i.e setting to 9 will
        subdivide the image into a 3x3 grid when comparing to the previous image.
    :type num_segments: int
    :param debug: When set, draws a red box around each changed image segment to
        show the area that changed. This is obviously destructive for displaying
        the actual image, but intends to show where the tracked changes are.
    :type debug: boolean
    """

    def __init__(self, num_segments=4, debug=False):
        self.__debug = debug
        self.__n = int(sqrt(num_segments))
        assert num_segments >= 1 and num_segments == self.__n ** 2
        self.prev_image = None

    def redraw(self, image):
        """
        Calculates the difference from the previous image, returning a sequence of
        image sections and bounding boxes that changed since the previous image.

        .. note::
            the first redraw will always render the full frame.

        :param image: The image to render.
        :type image: PIL.Image.Image
        :returns: Yields a sequence of images and the bounding box for each segment difference
        :rtype: Generator[Tuple[PIL.Image.Image, Tuple[int, int, int, int]]]
        """
        image_width, image_height = image.size
        segment_width = int(image_width / self.__n)
        segment_height = int(image_height / self.__n)
        assert segment_width * self.__n == image_width, "Total segment width does not cover full image width"
        assert segment_height * self.__n == image_height, "Total segment height does not cover full image height"

        changes = 0

        # Force a full redraw on the first frame
        if self.prev_image is None:
            changes += 1
            yield image, (0, 0) + image.size

        else:
            for y in range(0, image_height, segment_height):
                for x in range(0, image_width, segment_width):
                    bounding_box = (x, y, x + segment_width, y + segment_height)
                    prev_segment = self.prev_image.crop(bounding_box)
                    curr_segment = image.crop(bounding_box)
                    segment_bounding_box = ImageChops.difference(prev_segment, curr_segment).getbbox()
                    if segment_bounding_box is not None:
                        changes += 1
                        segment_bounding_box_from_origin = (
                            x + segment_bounding_box[0],
                            y + segment_bounding_box[1],
                            x + segment_bounding_box[2],
                            y + segment_bounding_box[3]
                        )

                        image_delta = curr_segment.crop(segment_bounding_box)

                        if self.__debug:
                            w, h = image_delta.size
                            draw = ImageDraw.Draw(image_delta)
                            draw.rectangle((0, 0, w - 1, h - 1), outline="red")
                            del draw

                        yield image_delta, segment_bounding_box_from_origin

        if changes > 0:
            self.prev_image = image.copy()


class full_frame(object):
    """
    Always renders the full frame every time. This is slower than
    :py:class:`diff_to_previous` as there are generally more
    pixels to update on every render, but it has a more consistent render time.
    Not all display drivers may be able to use the differencing framebuffer, so
    this is provided as a drop-in replacement.
    """

    def __init__(self, **kwargs):
        """
        Accepts any args but does nothing
        """

    def redraw(self, image):
        """
        Yields the full image for every redraw.

        :param image: The image to render.
        :type image: PIL.Image.Image
        :returns: Yields a single tuple of an image and the bounding box for that image
        :rtype: Generator[Tuple[PIL.Image.Image, Tuple[int, int, int, int]]]
        """
        yield image, (0, 0) + image.size



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-18 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.render.canvas` class.
"""

import pytest

from PIL import Image

from luma.core.device import dummy
from luma.core.render import canvas

from helpers import get_reference_image, assert_identical_image, test_font


def test_canvas_background():
    img_path = get_reference_image('background.png')
    with open(get_reference_image('dither.png'), 'rb') as p1:
        with open(img_path, 'rb') as p2:
            bgnd = Image.open(p1)
            reference = Image.open(p2)
            device = dummy()

            with canvas(device, background=bgnd) as draw:
                draw.rectangle((20, 15, device.width - 20, device.height - 15),
                               fill='black', outline='white')
                draw.text((36, 22), 'Background', fill='white', font=test_font)
                draw.text((52, 32), 'Test', fill='white', font=test_font)

            assert_identical_image(reference, device.image, img_path)


def test_canvas_wrong_size():
    with pytest.raises(AssertionError):
        canvas(dummy(), background=Image.new('RGB', (23, 97)))



# -*- coding: utf-8 -*-
# Copyright (c) 2017-2022 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Test helpers.
"""

import platform
from pathlib import Path

import pytest
from PIL import ImageChops, ImageFont
from unittest.mock import mock_open


rpi_gpio_missing = f'RPi.GPIO is not supported on this platform: {platform.system()}'
spidev_missing = f'spidev is not supported on this platform: {platform.system()}'


def get_reference_file(fname):
    """
    Get absolute path for ``fname``.

    :param fname: Filename.
    :type fname: str or pathlib.Path
    :rtype: str
    """
    return str(Path(__file__).resolve().parent.joinpath('reference', fname))


def get_reference_image(fname):
    """
    :param fname: Filename.
    :type fname: str or pathlib.Path
    """
    return get_reference_file(Path('images').joinpath(fname))


def get_reference_font(fname, fsize=12):
    """
    :param fname: Filename of the font.
    :type fname: str or pathlib.Path
    """
    path = get_reference_file(Path('font').joinpath(fname))
    return ImageFont.truetype(path, fsize)


def get_reference_pillow_font(fname):
    """
    Load :py:class:`PIL.ImageFont` type font from provided fname

    :param fname: The name of the file that contains the PIL.ImageFont
    :type fname: str
    :rtype: :py:class:`PIL.ImageFont`
    """
    path = get_reference_file(Path('font').joinpath(fname))
    return ImageFont.load(path)


# font used in (most) tests
test_font = get_reference_pillow_font('courB08.pil')


def get_spidev():
    try:
        import spidev
        return spidev
    except ImportError:
        pytest.skip(spidev_missing)


def assert_identical_image(reference, target, img_path):
    """
    :param img_path: Location of image.
    :type img_path: str
    """
    bbox = ImageChops.difference(reference, target).getbbox()
    #target.save(img_path)
    assert bbox is None, f'{img_path} is not identical to generated image'


def i2c_error(path_name, err_no):
    expected_error = OSError()
    expected_error.errno = err_no
    expected_error.filename = path_name

    def fake_open(a, b):
        raise expected_error
    return fake_open


def fib(n):
    a, b = 0, 1
    for _ in range(n):
        yield a
        a, b = b, a + b


# Attribution: https://gist.github.com/adammartinez271828/137ae25d0b817da2509c1a96ba37fc56
def multi_mock_open(*file_contents):
    """Create a mock "open" that will mock open multiple files in sequence
    Args:
        *file_contents ([str]): a list of file contents to be returned by open
    Returns:
        (MagicMock) a mock opener that will return the contents of the first
            file when opened the first time, the second file when opened the
            second time, etc.
    """
    mock_files = [mock_open(read_data=content) for content in file_contents]
    mock_opener = mock_files[-1]
    mock_opener.side_effect = [mock_file.return_value for mock_file in mock_files]

    return mock_opener


def skip_unsupported_platform(err):
    pytest.skip(f'{type(err).__name__} ({str(err)})')


def _positional_args_list(mock):
    return [call[0] for call in mock.call_args_list]


def assert_only_cleans_whats_setup(gpio):
    setups = _positional_args_list(gpio.setup)
    cleanups = _positional_args_list(gpio.cleanup)
    for cleanup in cleanups:
        assert len(cleanup) > 0, 'calling gpio.cleanup without specifying pins cleans all pins'
    pins_set_up = {args[0] for args in setups}
    pins_clean = {args[0] for args in setups}
    assert pins_clean == pins_set_up, f"set pins {pins_set_up} but cleaned pins {pins_clean}"



# -*- coding: utf-8 -*-
# Copyright (c) 2017-2020 Richard Hull and contributors
# See LICENSE.rst for details.

import os
import atexit
from time import sleep
from itertools import islice
from PIL import Image

from luma.core import mixin
from luma.core.util import bytes_to_nibbles
from luma.core.framebuffer import diff_to_previous
import luma.core.const
from luma.core.interface.serial import i2c, noop

__all__ = ["linux_framebuffer"]


class device(mixin.capabilities):
    """
    Base class for display driver classes

    .. note::
        Direct use of the :func:`command` and :func:`data` methods are
        discouraged: Screen updates should be effected through the
        :func:`display` method, or preferably with the
        :class:`luma.core.render.canvas` context manager.
    """

    def __init__(self, const=None, serial_interface=None):
        self._const = const or luma.core.const.common
        self._serial_interface = serial_interface or i2c()

        def shutdown_hook():  # pragma: no cover
            try:
                self.cleanup()
            except:
                pass

        atexit.register(shutdown_hook)

    def command(self, *cmd):
        """
        Sends a command or sequence of commands through to the delegated
        serial interface.
        """
        self._serial_interface.command(*cmd)

    def data(self, data):
        """
        Sends a data byte or sequence of data bytes through to the delegated
        serial interface.
        """
        self._serial_interface.data(data)

    def show(self):
        """
        Sets the display mode ON, waking the device out of a prior
        low-power sleep mode.
        """
        self.command(self._const.DISPLAYON)

    def hide(self):
        """
        Switches the display mode OFF, putting the device in low-power
        sleep mode.
        """
        self.command(self._const.DISPLAYOFF)

    def contrast(self, level):
        """
        Switches the display contrast to the desired level, in the range
        0-255. Note that setting the level to a low (or zero) value will
        not necessarily dim the display to nearly off. In other words,
        this method is **NOT** suitable for fade-in/out animation.

        :param level: Desired contrast level in the range of 0-255.
        :type level: int
        """
        assert 0 <= level <= 255
        self.command(self._const.SETCONTRAST, level)

    def cleanup(self):
        """
        Attempt to switch the device off or put into low power mode (this
        helps prolong the life of the device), clear the screen and close
        resources associated with the underlying serial interface.

        If :py:attr:`persist` is ``True``, the device will not be switched off.

        This is a managed function, which is called when the python processs
        is being shutdown, so shouldn't usually need be called directly in
        application code.
        """
        if not self.persist:
            self.hide()
            self.clear()
        self._serial_interface.cleanup()


class parallel_device(device):
    """
    Wrapper class to manage communications with devices that can operate in
    four or eight bit modes.

    .. note::
        parallel_devices require specific timings which are managed by using
        ``time.sleep`` to cause the process to block for small amounts of time.
        If your application is especially time sensitive, consider running the
        drivers in a separate thread.

    .. versionadded:: 1.16.0
    """

    def __init__(self, const=None, serial_interface=None, exec_time=None, **kwargs):
        super(parallel_device, self).__init__(const, serial_interface)

        self._exec_time = exec_time if exec_time is not None else \
            serial_interface._pulse_time if hasattr(serial_interface, '_pulse_time') \
            else 0
        self._bitmode = serial_interface._bitmode if hasattr(serial_interface, '_bitmode') else 4
        assert self._bitmode in (4, 8), f'Bit mode {self._bitmode} is invalid.  It can only be 4 or 8'

    def command(self, *cmd, exec_time=None, only_low_bits=False):
        """
        Sends a command or sequence of commands through to the serial interface.
        If operating in four bit mode, expands each command from one byte
        values (8 bits) to two nibble values (4 bits each)

        :param cmd: A spread of commands.
        :type cmd: int
        :param exec_time: Amount of time to wait for the command to finish
            execution.  If not provided, the device default will be used instead
        :type exec_time: float
        :param only_low_bits: If ``True``, only the lowest four bits of the command
            will be sent.  This is necessary on some devices during initialization
        :type only_low_bits: bool
        """
        cmd = cmd if (self._bitmode == 8 or only_low_bits) else \
            bytes_to_nibbles(cmd)
        super(parallel_device, self).command(*cmd)
        sleep(exec_time or self._exec_time)

    def data(self, data):
        """
        Sends a sequence of bytes through to the serial interface.
        If operating in four bit mode, expands each byte from a single
        value (8 bits) to two nibble values (4 bits each)

        :param data: a sequence of bytes to send to the display
        :type data: list
        """
        data = data if self._bitmode == 8 else \
            bytes_to_nibbles(data)
        super(parallel_device, self).data(data)


class dummy(device):
    """
    Pseudo-device that acts like a physical display, except that it does nothing
    other than retain a copy of the displayed image. It is mostly useful for
    testing. Supports 24-bit color depth.
    """

    def __init__(self, width=128, height=64, rotate=0, mode="RGB", **kwargs):
        super(dummy, self).__init__(serial_interface=noop())
        self.capabilities(width, height, rotate, mode)
        self.image = None

    def display(self, image):
        """
        Takes a :py:mod:`PIL.Image` and makes a copy of it for later
        use/inspection.

        :param image: Image to display.
        :type image: PIL.Image.Image
        """
        assert image.size == self.size

        self.image = self.preprocess(image).copy()


class linux_framebuffer(device):
    """
    Pseudo-device that acts like a physical display, except that it renders
    to a Linux framebuffer device at /dev/fbN (where N=0, 1, ...). This is specifically
    targetted to allow the luma classes to be used on higher-resolution displays that
    leverage kernel-based display drivers.

    .. note:
        Currently only supports 16-bit RGB, 24-bit RGB/BGR and 32-bit RGBA/BGRA color depths.

    :param device: the Linux framebuffer device (e.g. `/dev/fb0`). If no device
        is given, the device is determined from the `FRAMEBUFFER` environmental
        variable instead. See https://www.kernel.org/doc/html/latest/fb/framebuffer.html
        for more details.
    :param framebuffer: Framebuffer rendering strategy, currently instances of
        ``diff_to_previous`` (default, if not specified) or ``full_frame``.
    :param bgr: Set to ``True`` if device pixels are BGR order (rather than RGB). Note:
        this flag is currently supported on 24 and 32-bit color depth devices only.

    .. versionadded:: 2.0.0
    """

    def __init__(self, device=None, framebuffer=None, bgr=False, **kwargs):
        super(linux_framebuffer, self).__init__(serial_interface=noop())
        self.id = self.__get_display_id(device)
        (width, height) = self.__config("virtual_size")
        self.bits_per_pixel = next(self.__config("bits_per_pixel"))

        # Lookup table of (target bit-depth, bgr) -> function, where the
        # function takes an RGB image and converts it into a stream of
        # bytes for the target bit-depth and RGB/BGR orientation.
        image_converters = {
            (16, False): self.__toRGB565,
            (24, False): self.__toRGB,
            (24, True): self.__toBGR,
            (32, False): self.__toRGBA,
            (32, True): self.__toBGRA,
        }
        assert (self.bits_per_pixel, bgr) in image_converters, f"Unsupported bit-depth: {self.bits_per_pixel}"
        self.__image_converter = image_converters[(self.bits_per_pixel, bgr)]

        self.framebuffer = framebuffer or diff_to_previous(num_segments=16)
        self.capabilities(width, height, rotate=0, mode="RGB")

        # This file handle is closed in self.cleanup() (usually invoked
        # automatically via a registered `atexit` hook)
        self.__file_handle = open(f"/dev/fb{self.id}", "wb")

    def __get_display_id(self, device):
        """
        Extract the display-id from the device which is usually referred in
        the form `/dev/fbN` where N is numeric. If no device is given, defer
        to the FRAMEBUFFER environmental variable.

        See https://www.kernel.org/doc/html/latest/fb/framebuffer.html for more details.
        """
        if device is None:
            device = os.environ.get("FRAMEBUFFER", "/dev/fb0")

        if device.startswith("/dev/fb"):
            return int(device[7:])

        raise luma.core.error.DeviceNotFoundError(
            f"Invalid/unsupported framebuffer: {device}"
        )

    def __config(self, section):
        path = f"/sys/class/graphics/fb{self.id}/{section}"
        with open(path, "r") as fp:
            for value in fp.read().strip().split(","):
                if value:
                    yield int(value)

    def __toRGB565(self, image):
        for r, g, b in image.getdata():
            yield g << 3 & 0xE0 | b >> 3
            yield r & 0xF8 | g >> 5

    def __toRGB(self, image):
        return iter(image.tobytes())

    def __toBGR(self, image):
        r, g, b = image.split()
        return iter(Image.merge("RGB", (b, g, r)).tobytes())

    def __toRGBA(self, image):
        return iter(image.convert("RGBA").tobytes())

    def __toBGRA(self, image):
        r, g, b = image.split()
        return iter(Image.merge("RGB", (b, g, r)).convert("RGBA").tobytes())

    def cleanup(self):
        super(linux_framebuffer, self).cleanup()
        self.__file_handle.close()

    def display(self, image):
        """
        Takes a :py:mod:`PIL.Image` and converts it for consumption on the
        given /dev/fbX framebuffer device.

        :param image: Image to display.
        :type image: PIL.Image.Image
        """
        assert image.mode == self.mode
        assert image.size == self.size

        image = self.preprocess(image)
        file_handle = self.__file_handle

        bytes_per_pixel = self.bits_per_pixel // 8
        image_bytes_per_row = self.width * bytes_per_pixel

        for image, bounding_box in self.framebuffer.redraw(image):
            left, top, right, bottom = bounding_box
            segment_bytes_per_row = (right - left) * bytes_per_pixel
            left_offset = left * bytes_per_pixel
            generator = self.__image_converter(image)
            for row_offset in range(left_offset + top * image_bytes_per_row,
                                    left_offset + bottom * image_bytes_per_row,
                                    image_bytes_per_row):
                file_handle.seek(row_offset)
                file_handle.write(bytes(islice(generator, segment_bytes_per_row)))

        file_handle.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Richard Hull and contributors
# See LICENSE.rst for details.

"""
Tests for the :py:class:`luma.core.interface.serial.ftdi_spi` class.
"""

from unittest.mock import Mock, call, patch
from luma.core.interface.serial import ftdi_spi
from helpers import fib, assert_only_cleans_whats_setup


@patch('pyftdi.spi.SpiController')
def test_init(mock_controller):
    gpio = Mock()
    gpio.width = 8
    instance = Mock()
    instance.get_port = Mock()
    instance.get_gpio = Mock(return_value=gpio)
    mock_controller.side_effect = [instance]

    ftdi_spi(device='ftdi://dummy', bus_speed_hz=16000000, gpio_CS=3, gpio_DC=5, gpio_RST=6)
    mock_controller.assert_called_with(cs_count=1)
    instance.configure.assert_called_with('ftdi://dummy')
    instance.get_port.assert_called_with(cs=0, freq=16000000, mode=0)
    instance.get_gpio.assert_called_with()
    gpio.set_direction.assert_called_with(0x60, 0x60)


@patch('pyftdi.spi.SpiController')
def test_command(mock_controller):
    cmds = [3, 1, 4, 2]
    gpio = Mock()
    gpio.width = 8
    port = Mock()
    instance = Mock()
    instance.get_port = Mock(return_value=port)
    instance.get_gpio = Mock(return_value=gpio)
    mock_controller.side_effect = [instance]

    serial = ftdi_spi(device='ftdi://dummy', bus_speed_hz=16000000, gpio_CS=3, gpio_DC=5, gpio_RST=6)
    serial.command(*cmds)
    gpio.write.assert_has_calls([call(0x00), call(0x40), call(0x40)])
    port.write.assert_called_once_with(cmds)


@patch('pyftdi.spi.SpiController')
def test_data(mock_controller):
    data = list(fib(100))
    gpio = Mock()
    gpio.width = 8
    port = Mock()
    instance = Mock()
    instance.get_port = Mock(return_value=port)
    instance.get_gpio = Mock(return_value=gpio)
    mock_controller.side_effect = [instance]

    serial = ftdi_spi(device='ftdi://dummy', bus_speed_hz=16000000, gpio_CS=3, gpio_DC=5, gpio_RST=6)
    serial.data(data)
    gpio.write.assert_has_calls([call(0x00), call(0x40), call(0x60)])
    port.write.assert_called_once_with(data)


@patch('pyftdi.spi.SpiController')
def test_cleanup(mock_controller):
    gpio = Mock()
    gpio.width = 8
    port = Mock()
    instance = Mock()
    instance.get_port = Mock(return_value=port)
    instance.get_gpio = Mock(return_value=gpio)
    mock_controller.side_effect = [instance]

    serial = ftdi_spi(device='ftdi://dummy', bus_speed_hz=16000000, gpio_CS=3, gpio_DC=5, gpio_RST=6)
    serial.cleanup()
    instance.terminate.assert_called_once_with()
    assert_only_cleans_whats_setup(gpio)
